use derive_more::{Add, AddAssign, Display, Sub, SubAssign};
use crate::quantity::{Quantity, Value};

const VOLT_MV_RATIO: u16 = 1_000;

// ---------------------------------------------------------------------------------------------------------------------
// ----- Volts ------------------------
// ---------------------------------------------------------------------------------------------------------------------
#[derive(Copy, Clone, PartialEq, PartialOrd, Add, AddAssign, Sub, SubAssign, Display)]
#[display(fmt = "{} {}", _0, "Self::symbol()")]
#[repr(transparent)]
pub struct Volts<V: Value>(pub V);

impl <V: Value> Quantity<V> for Volts<V> {
    #[inline(always)]
    fn value(self) -> V {
        self.0
    }

    #[inline(always)]
    fn symbol() -> &'static str {
        "V"
    }
}

impl<V: Value> Volts<V> {
    #[inline]
    pub fn to_millivolts(self) -> MilliVolts<V> {
        let multiplier = V::from_u16(VOLT_MV_RATIO).unwrap();
        MilliVolts(self.0 * multiplier)
    }
}

// ---------------------------------------------------------------------------------------------------------------------
// ----- MilliVolts ------------------------
// ---------------------------------------------------------------------------------------------------------------------
#[derive(Copy, Clone, PartialEq, PartialOrd, Add, AddAssign, Sub, SubAssign, Display)]
#[display(fmt = "{} {}", _0, "Self::symbol()")]
#[repr(transparent)]
pub struct MilliVolts<V: Value>(pub V);

impl <V: Value> Quantity<V> for MilliVolts<V> {
    #[inline(always)]
    fn value(self) -> V {
        self.0
    }

    #[inline(always)]
    fn symbol() -> &'static str {
        "mV"
    }
}

impl<V: Value> MilliVolts<V> {
    pub fn to_volts(self) -> Volts<V> {
        let divisor = V::from_u16(VOLT_MV_RATIO).unwrap();
        Volts(self.0 / divisor)
    }
}

// ---------------------------------------------------------------------------------------------------------------------
// ----- Tests ------------------------
// ---------------------------------------------------------------------------------------------------------------------
#[cfg(test)]
mod tests {
    use float_cmp::assert_approx_eq;
    use super::*;
    
    #[test]
    fn convert_u32() {
        let volts: Volts<u32> = 3.volts();
        let millivolts: MilliVolts<u32> = 3_000.millivolts();
        
        assert_eq!(volts.to_millivolts().0, millivolts.0);
        assert_eq!(millivolts.to_volts().0, volts.0);
    }
    
    #[test]
    fn convert_f64() {
        let volts: Volts<f64> = 3.0.volts();
        let millivolts: MilliVolts<f64> = 3_000.0.millivolts();
        
        assert_approx_eq!(f64, volts.to_millivolts().0, millivolts.0);
        assert_approx_eq!(f64, millivolts.to_volts().0, volts.0);
    }
}